@@ -427,7 +427,8 @@ takeCurrentFrame = do
     getByteString (fromIntegral size - 2)
 
 putFrame :: JpgFrame -> Put
-putFrame (JpgAdobeAPP14 _adobe) = return ()
+putFrame (JpgAdobeAPP14 adobe) = 
+    put (JpgAppSegment 14) >> putWord16be 14 >> put adobe
 putFrame (JpgJFIF jfif) =
     put (JpgAppSegment 0) >> putWord16be (14+2) >> put jfif
 putFrame (JpgExif _exif) =
@@ -443,8 +444,16 @@ putFrame (JpgHuffmanTable tables) =
     put JpgHuffmanTableMarker >> put (TableList $ map fst tables)
 putFrame (JpgIntervalRestart size) =
     put JpgRestartInterval >> put (RestartInterval size)
-putFrame (JpgScanBlob hdr blob) =
-    put JpgStartOfScan >> put hdr >> putLazyByteString blob
+putFrame (JpgScanBlob hdr blob) = do
+    put JpgStartOfScan
+    put hdr
+    putLazyByteString blob
+    putWord8 0 -- AKA the libjpeg pleaser, for some unknown reason
+               -- libjpeg raise a warning "invalid end of data segment".
+               -- so dumbly pad, seems to make the warning go away (and
+               -- I don't want to invest that much time looking for that).
+               --
+               -- Ok this is a crummy fix....
 putFrame (JpgScans kind hdr) =
     put kind >> put hdr
 
